/**
 * \file pappsomspp/psm/peptidespectrummatch.cpp
 * \date 2/4/2015
 * \author Olivier Langella
 * \brief find peaks matching between ions and spectrum
 */

/*******************************************************************************
 * Copyright (c) 2015 Olivier Langella <Olivier.Langella@moulon.inra.fr>.
 *
 * This file is part of the PAPPSOms++ library.
 *
 *     PAPPSOms++ is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     PAPPSOms++ is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with PAPPSOms++.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contributors:
 *     Olivier Langella <Olivier.Langella@moulon.inra.fr> - initial API and implementation
 ******************************************************************************/


#include <list>
#include "peakionmatch.h"
#include "peptidespectrummatch.h"
#include "../pappsoexception.h"
#include "../peptide/peptidefragment.h"
#include "../peptide/peptidefragmentionlistbase.h"

namespace pappso {


void PeptideSpectrumMatch::privMatchIonList(const Spectrum & spectrum, const PeptideFragmentIonListBase & fragmentIonList, unsigned int max_charge, std::list<PeptideIon> ion_type_list) {

    std::list<Peak> peak_list(spectrum.begin(), spectrum.end());

    for (auto ion_type: ion_type_list) {
        auto ion_list = fragmentIonList.getPeptideFragmentIonSp(ion_type);

        for (unsigned int charge=1; charge <= max_charge; charge++) {
            for (auto && ion : ion_list) {
                std::list<Peak>::iterator it_peak = getBestPeakIterator (peak_list, ion, charge);
                if (it_peak != peak_list.end()) {
                    _peak_ion_match_list.push_back(PeakIonMatch(*it_peak, ion, charge));
                    peak_list.erase(it_peak);
                }
            }
        }
    }


}

PeptideSpectrumMatch::PeptideSpectrumMatch(const Spectrum & spectrum,std::vector<PeptideFragmentIonSp> & v_peptide_fragment_ion,  std::vector<unsigned int> & v_peptide_fragment_ion_charge, PrecisionP precision): _precision(precision) {
    qDebug() << "PeptideSpectrumMatch::PeptideSpectrumMatch vector list based begin";
    // throw PappsoException(QObject::tr("v_peptideIsotopeList.size() %1 != v_peptideIonList.size() %2").arg(v_peptideIsotopeList.size()).arg(v_peptideIonList.size()));
    if (v_peptide_fragment_ion.size() != v_peptide_fragment_ion_charge.size()) {
        throw PappsoException(QObject::tr("v_peptide_fragment_ion.size() != v_peptide_fragment_ion_charge.size() %2").arg(v_peptide_fragment_ion.size()).arg(v_peptide_fragment_ion_charge.size()));
    }


    auto ionIt = v_peptide_fragment_ion.begin();
    auto chargeIt = v_peptide_fragment_ion_charge.begin();
    std::list<Peak> peak_list(spectrum.begin(), spectrum.end());

    while (ionIt != v_peptide_fragment_ion.end()) {
        std::list<Peak>::iterator it_peak = getBestPeakIterator (peak_list, *ionIt,*chargeIt);
        if (it_peak != peak_list.end()) {
            _peak_ion_match_list.push_back(PeakIonMatch(*it_peak, *ionIt, *chargeIt));
            peak_list.erase(it_peak);
        }

        ionIt++;
        chargeIt++;
    }
    qDebug() << "PeptideSpectrumMatch::PeptideSpectrumMatch vector list based end";
}


PeptideSpectrumMatch::PeptideSpectrumMatch(const Spectrum & spectrum, const PeptideFragmentIonListBase & fragmentIonList, unsigned int max_charge, PrecisionP precision, std::list<PeptideIon> ion_type_list): _precision(precision) {
    privMatchIonList(spectrum, fragmentIonList, max_charge, ion_type_list);


}

PeptideSpectrumMatch::PeptideSpectrumMatch(const Spectrum & spectrum, pappso::PeptideSp peptideSp, unsigned int max_charge, PrecisionP precision, std::list<PeptideIon> ion_type_list): _precision(precision)
{
    PeptideFragmentIonListBase fragmentIonList(peptideSp, ion_type_list);
    privMatchIonList(spectrum, fragmentIonList, max_charge, ion_type_list);
}


PeptideSpectrumMatch::PeptideSpectrumMatch(const PeptideSpectrumMatch& other):_precision(other._precision), _peak_ion_match_list(other._peak_ion_match_list)
{

}

std::list<Peak>::iterator PeptideSpectrumMatch::getBestPeakIterator (std::list<Peak> & peak_list, const PeptideFragmentIonSp & ion,unsigned int charge) const {
    std::list<Peak>::iterator itpeak = peak_list.begin();
    std::list<Peak>::iterator itend = peak_list.end();
    std::list<Peak>::iterator itselect = peak_list.end();

    pappso_double best_intensity = 0;

    while (itpeak != itend) {
        if (ion.get()->matchPeak(_precision, itpeak->mz, charge)) {
            if (itpeak->intensity > best_intensity) {
                best_intensity = itpeak->intensity;
                itselect = itpeak;
            }
        }
        itpeak++;
    }
    return(itselect);
}

PeptideSpectrumMatch::~PeptideSpectrumMatch()
{

}

bool PeptideSpectrumMatch::contains(PeptideFragmentIonSp peptideFragmentIonSp, unsigned int z) const {
    auto it = _peak_ion_match_list.begin();
    while (it != _peak_ion_match_list.end()) {
        const PeakIonMatch & ref = *it;
        if ((ref.getCharge() == z) && (ref.getPeptideFragmentIonSp().get() == peptideFragmentIonSp.get())) {
            return true;
        }
        it++;
    }
    return false;
}

}
